"""Code is adapted from https://github.com/MIT-AI-Accelerator/neurips-2020-sevir. Their license is MIT License."""

import os
from typing import List, Union, Dict, Sequence
from math import ceil
import numpy as np
import numpy.random as nprand
import datetime
import pandas as pd
import h5py
from ...utils.layout import change_layout_np, change_layout_torch
import torch
from torch.nn.functional import avg_pool2d


# SEVIR Dataset constants
SEVIR_DATA_TYPES = ['vis', 'ir069', 'ir107', 'vil', 'lght']
SEVIR_RAW_DTYPES = {'vis': np.int16,
                    'ir069': np.int16,
                    'ir107': np.int16,
                    'vil': np.uint8,
                    'lght': np.int16}
LIGHTING_FRAME_TIMES = np.arange(- 120.0, 125.0, 5) * 60
SEVIR_DATA_SHAPE = {'lght': (48, 48), }
PREPROCESS_SCALE_SEVIR = {'vis': 1,  # Not utilized in original paper
                          'ir069': 1 / 1174.68,
                          'ir107': 1 / 2562.43,
                          'vil': 1 / 47.54,
                          'lght': 1 / 0.60517}
PREPROCESS_OFFSET_SEVIR = {'vis': 0,  # Not utilized in original paper
                           'ir069': 3683.58,
                           'ir107': 1552.80,
                           'vil': - 33.44,
                           'lght': - 0.02990}
PREPROCESS_SCALE_01 = {'vis': 1,
                       'ir069': 1,
                       'ir107': 1,
                       'vil': 1 / 255,  # currently the only one implemented
                       'lght': 1}
PREPROCESS_OFFSET_01 = {'vis': 0,
                        'ir069': 0,
                        'ir107': 0,
                        'vil': 0,  # currently the only one implemented
                        'lght': 0}
# # default configs
from ...config import cfg
# sevir
SEVIR_ROOT_DIR = os.path.join(cfg.datasets_dir, "sevir")
SEVIR_CATALOG = os.path.join(SEVIR_ROOT_DIR, "CATALOG.csv")
SEVIR_DATA_DIR = os.path.join(SEVIR_ROOT_DIR, "data")
SEVIR_RAW_SEQ_LEN = 49
# sevir-lr
SEVIR_LR_ROOT_DIR = os.path.join(cfg.datasets_dir, "sevir_lr")
SEVIR_LR_CATALOG = os.path.join(SEVIR_LR_ROOT_DIR, "CATALOG.csv")
SEVIR_LR_DATA_DIR = os.path.join(SEVIR_LR_ROOT_DIR, "data")
SEVIR_LR_RAW_SEQ_LEN = 25


class SEVIRDataLoader:
    r"""
    DataLoader that loads SEVIR sequences, and spilts each event
    into segments according to specified sequence length.

    Event Frames:
        [-----------------------raw_seq_len----------------------]
        [-----seq_len-----]
        <--stride-->[-----seq_len-----]
                    <--stride-->[-----seq_len-----]
                                        ...
    """
    def __init__(self,
                 data_types: Sequence[str] = None,
                 seq_len: int = 49,
                 raw_seq_len: int = 49,
                 sample_mode: str = 'sequent',
                 stride: int = 12,
                 batch_size: int = 1,
                 layout: str = 'NHWT',
                 num_shard: int = 1,
                 rank: int = 0,
                 split_mode: str = "uneven",
                 sevir_catalog: Union[str, pd.DataFrame] = None,
                 sevir_data_dir: str = None,
                 start_date: datetime.datetime = None,
                 end_date: datetime.datetime = None,
                 datetime_filter=None,
                 catalog_filter='default',
                 shuffle: bool = False,
                 shuffle_seed: int = 1,
                 output_type=np.float32,
                 preprocess: bool = True,
                 rescale_method: str = '01',
                 downsample_dict: Dict[str, Sequence[int]] = None,
                 verbose: bool = False):
        r"""
        Parameters
        ----------
        data_types
            A subset of SEVIR_DATA_TYPES.
        seq_len
            The length of the data sequences. Should be smaller than the max length raw_seq_len.
        raw_seq_len
            The length of the raw data sequences.
        sample_mode
            'random' or 'sequent'
        stride
            Useful when sample_mode == 'sequent'
            stride must not be smaller than out_len to prevent data leakage in testing.
        batch_size
            Number of sequences in one batch.
        layout
            str: consists of batch_size 'N', seq_len 'T', channel 'C', height 'H', width 'W'
            The layout of sampled data. Raw data layout is 'NHWT'.
            valid layout: 'NHWT', 'NTHW', 'NTCHW', 'TNHW', 'TNCHW'.
        num_shard
            Split the whole dataset into num_shard parts for distributed training.
        rank
            Rank of the current process within num_shard.
        split_mode: str
            if 'ceil', all `num_shard` dataloaders have the same length = ceil(total_len / num_shard).
            Different dataloaders may have some duplicated data batches, if the total size of datasets is not divided by num_shard.
            if 'floor', all `num_shard` dataloaders have the same length = floor(total_len / num_shard).
            The last several data batches may be wasted, if the total size of datasets is not divided by num_shard.
            if 'uneven', the last datasets has larger length when the total length is not divided by num_shard.
            The uneven split leads to synchronization error in dist.all_reduce() or dist.barrier().
            See related issue: https://github.com/pytorch/pytorch/issues/33148
            Notice: this also affects the behavior of `self.use_up`.
        sevir_catalog
            Name of SEVIR catalog CSV file.
        sevir_data_dir
            Directory path to SEVIR data.
        start_date
            Start time of SEVIR samples to generate.
        end_date
            End time of SEVIR samples to generate.
        datetime_filter
            function
            Mask function applied to time_utc column of catalog (return true to keep the row).
            Pass function of the form   lambda t : COND(t)
            Example:  lambda t: np.logical_and(t.dt.hour>=13,t.dt.hour<=21)  # Generate only day-time events
        catalog_filter
            function or None or 'default'
            Mask function applied to entire catalog dataframe (return true to keep row).
            Pass function of the form lambda catalog:  COND(catalog)
            Example:  lambda c:  [s[0]=='S' for s in c.id]   # Generate only the 'S' events
        shuffle
            bool, If True, data samples are shuffled before each epoch.
        shuffle_seed
            int, Seed to use for shuffling.
        output_type
            np.dtype, dtype of generated tensors
        preprocess
            bool, If True, self.preprocess_data_dict(data_dict) is called before each sample generated
        downsample_dict:
            dict, downsample_dict.keys() == data_types. downsample_dict[key] is a Sequence of (t_factor, h_factor, w_factor),
            representing the downsampling factors of all dimensions.
        verbose
            bool, verbose when opening raw data files
        """
        super(SEVIRDataLoader, self).__init__()
        if sevir_catalog is None:
            sevir_catalog = SEVIR_CATALOG
        if sevir_data_dir is None:
            sevir_data_dir = SEVIR_DATA_DIR
        if data_types is None:
            data_types = SEVIR_DATA_TYPES
        else:
            assert set(data_types).issubset(SEVIR_DATA_TYPES)

        # configs which should not be modified
        self._dtypes = SEVIR_RAW_DTYPES
        self.lght_frame_times = LIGHTING_FRAME_TIMES
        self.data_shape = SEVIR_DATA_SHAPE

        self.raw_seq_len = raw_seq_len
        assert seq_len <= self.raw_seq_len, f'seq_len must not be larger than raw_seq_len = {raw_seq_len}, got {seq_len}.'
        self.seq_len = seq_len
        assert sample_mode in ['random', 'sequent'], f'Invalid sample_mode = {sample_mode}, must be \'random\' or \'sequent\'.'
        self.sample_mode = sample_mode
        self.stride = stride
        self.batch_size = batch_size
        valid_layout = ('NHWT', 'NTHW', 'NTCHW', 'NTHWC', 'TNHW', 'TNCHW')
        if layout not in valid_layout:
            raise ValueError(f'Invalid layout = {layout}! Must be one of {valid_layout}.')
        self.layout = layout
        self.num_shard = num_shard
        self.rank = rank
        valid_split_mode = ('ceil', 'floor', 'uneven')
        if split_mode not in valid_split_mode:
            raise ValueError(f'Invalid split_mode: {split_mode}! Must be one of {valid_split_mode}.')
        self.split_mode = split_mode
        self._samples = None
        self._hdf_files = {}
        self.data_types = data_types
        if isinstance(sevir_catalog, str):
            self.catalog = pd.read_csv(sevir_catalog, parse_dates=['time_utc'], low_memory=False)
        else:
            self.catalog = sevir_catalog
        self.sevir_data_dir = sevir_data_dir
        self.datetime_filter = datetime_filter
        self.catalog_filter = catalog_filter
        self.start_date = start_date
        self.end_date = end_date
        self.shuffle = shuffle
        self.shuffle_seed = int(shuffle_seed)
        self.output_type = output_type
        self.preprocess = preprocess
        self.downsample_dict = downsample_dict
        self.rescale_method = rescale_method
        self.verbose = verbose

        if self.start_date is not None:
            self.catalog = self.catalog[self.catalog.time_utc > self.start_date]
        if self.end_date is not None:
            self.catalog = self.catalog[self.catalog.time_utc <= self.end_date]
        if self.datetime_filter:
            self.catalog = self.catalog[self.datetime_filter(self.catalog.time_utc)]

        if self.catalog_filter is not None:
            if self.catalog_filter == 'default':
                self.catalog_filter = lambda c: c.pct_missing == 0
            self.catalog = self.catalog[self.catalog_filter(self.catalog)]

        self._compute_samples()
        self._open_files(verbose=self.verbose)
        self.reset()

    def _compute_samples(self):
        """
        Computes the list of samples in catalog to be used. This sets self._samples
        """
        # locate all events containing colocated data_types
        imgt = self.data_types
        imgts = set(imgt)
        filtcat = self.catalog[ np.logical_or.reduce([self.catalog.img_type==i for i in imgt]) ]
        # remove rows missing one or more requested img_types
        filtcat = filtcat.groupby('id').filter(lambda x: imgts.issubset(set(x['img_type'])))
        # If there are repeated IDs, remove them (this is a bug in SEVIR)
        # TODO: is it necessary to keep one of them instead of deleting them all
        filtcat = filtcat.groupby('id').filter(lambda x: x.shape[0]==len(imgt))
        self._samples = filtcat.groupby('id').apply(lambda df: self._df_to_series(df,imgt) )
        if self.shuffle:
            self.shuffle_samples()

    def shuffle_samples(self):
        self._samples = self._samples.sample(frac=1, random_state=self.shuffle_seed)

    def _df_to_series(self, df, imgt):
        d = {}
        df = df.set_index('img_type')
        for i in imgt:
            s = df.loc[i]
            idx = s.file_index if i != 'lght' else s.id
            d.update({f'{i}_filename': [s.file_name],
                      f'{i}_index': [idx]})

        return pd.DataFrame(d)

    def _open_files(self, verbose=True):
        """
        Opens HDF files
        """
        imgt = self.data_types
        hdf_filenames = []
        for t in imgt:
            hdf_filenames += list(np.unique( self._samples[f'{t}_filename'].values ))
        self._hdf_files = {}
        for f in hdf_filenames:
            if verbose:
                print('Opening HDF5 file for reading', f)
            self._hdf_files[f] = h5py.File(self.sevir_data_dir + '/' + f, 'r')

    def close(self):
        """
        Closes all open file handles
        """
        for f in self._hdf_files:
            self._hdf_files[f].close()
        self._hdf_files = {}

    @property
    def num_seq_per_event(self):
        return 1 + (self.raw_seq_len - self.seq_len) // self.stride

    @property
    def total_num_seq(self):
        """
        The total number of sequences within each shard.
        Notice that it is not the product of `self.num_seq_per_event` and `self.total_num_event`.
        """
        return int(self.num_seq_per_event * self.num_event)

    @property
    def total_num_event(self):
        """
        The total number of events in the whole dataset, before split into different shards.
        """
        return int(self._samples.shape[0])

    @property
    def start_event_idx(self):
        """
        The event idx used in certain rank should satisfy event_idx >= start_event_idx
        """
        return self.total_num_event // self.num_shard * self.rank

    @property
    def end_event_idx(self):
        """
        The event idx used in certain rank should satisfy event_idx < end_event_idx

        """
        if self.split_mode == 'ceil':
            _last_start_event_idx = self.total_num_event // self.num_shard * (self.num_shard - 1)
            _num_event = self.total_num_event - _last_start_event_idx
            return self.start_event_idx + _num_event
        elif self.split_mode == 'floor':
            return self.total_num_event // self.num_shard * (self.rank + 1)
        else:  # self.split_mode == 'uneven':
            if self.rank == self.num_shard - 1:  # the last process
                return self.total_num_event
            else:
                return self.total_num_event // self.num_shard * (self.rank + 1)

    @property
    def num_event(self):
        """
        The number of events split into each rank
        """
        return self.end_event_idx - self.start_event_idx

    def _read_data(self, row, data):
        """
        Iteratively read data into data dict. Finally data[imgt] gets shape (batch_size, height, width, raw_seq_len).

        Parameters
        ----------
        row
            A series with fields IMGTYPE_filename, IMGTYPE_index, IMGTYPE_time_index.
        data
            Dict, data[imgt] is a data tensor with shape = (tmp_batch_size, height, width, raw_seq_len).

        Returns
        -------
        data
            Updated data. Updated shape = (tmp_batch_size + 1, height, width, raw_seq_len).
        """
        imgtyps = np.unique([x.split('_')[0] for x in list(row.keys())])
        for t in imgtyps:
            fname = row[f'{t}_filename']
            idx = row[f'{t}_index']
            t_slice = slice(0, None)
            # Need to bin lght counts into grid
            if t == 'lght':
                lght_data = self._hdf_files[fname][idx][:]
                data_i = self._lght_to_grid(lght_data, t_slice)
            else:
                data_i = self._hdf_files[fname][t][idx:idx + 1, :, :, t_slice]
            data[t] = np.concatenate((data[t], data_i), axis=0) if (t in data) else data_i

        return data

    def _lght_to_grid(self, data, t_slice=slice(0, None)):
        """
        Converts Nx5 lightning data matrix into a 2D grid of pixel counts
        """
        # out_size = (48,48,len(self.lght_frame_times)-1) if isinstance(t_slice,(slice,)) else (48,48)
        out_size = (*self.data_shape['lght'], len(self.lght_frame_times)) if t_slice.stop is None else (*self.data_shape['lght'], 1)
        if data.shape[0] == 0:
            return np.zeros((1,) + out_size, dtype=np.float32)

        # filter out points outside the grid
        x, y = data[:, 3], data[:, 4]
        m = np.logical_and.reduce([x >= 0, x < out_size[0], y >= 0, y < out_size[1]])
        data = data[m, :]
        if data.shape[0] == 0:
            return np.zeros((1,) + out_size, dtype=np.float32)

        # Filter/separate times
        t = data[:, 0]
        if t_slice.stop is not None:  # select only one time bin
            if t_slice.stop > 0:
                if t_slice.stop < len(self.lght_frame_times):
                    tm = np.logical_and(t >= self.lght_frame_times[t_slice.stop - 1],
                                        t < self.lght_frame_times[t_slice.stop])
                else:
                    tm = t >= self.lght_frame_times[-1]
            else:  # special case:  frame 0 uses lght from frame 1
                tm = np.logical_and(t >= self.lght_frame_times[0], t < self.lght_frame_times[1])
            # tm=np.logical_and( (t>=FRAME_TIMES[t_slice],t<FRAME_TIMES[t_slice+1]) )

            data = data[tm, :]
            z = np.zeros(data.shape[0], dtype=np.int64)
        else:  # compute z coordinate based on bin location times
            z = np.digitize(t, self.lght_frame_times) - 1
            z[z == -1] = 0  # special case:  frame 0 uses lght from frame 1

        x = data[:, 3].astype(np.int64)
        y = data[:, 4].astype(np.int64)

        k = np.ravel_multi_index(np.array([y, x, z]), out_size)
        n = np.bincount(k, minlength=np.prod(out_size))
        return np.reshape(n, out_size).astype(np.int16)[np.newaxis, :]

    def _old_save_downsampled_dataset(self, save_dir, downsample_dict, verbose=True):
        """
        This method does not save .h5 dataset correctly. There are some batches missed due to unknown error.
        E.g., the first converted .h5 file `SEVIR_VIL_RANDOMEVENTS_2017_0501_0831.h5` only has batch_dim = 1414,
        while it should be 1440 in the original .h5 file.
        """
        import os
        from skimage.measure import block_reduce
        assert not os.path.exists(save_dir), f"save_dir {save_dir} already exists!"
        os.makedirs(save_dir)
        sample_counter = 0
        for index, row in self._samples.iterrows():
            if verbose:
                print(f"Downsampling {sample_counter}-th data item.", end='\r')
            for data_type in self.data_types:
                fname = row[f'{data_type}_filename']
                idx = row[f'{data_type}_index']
                t_slice = slice(0, None)
                if data_type == 'lght':
                    lght_data = self._hdf_files[fname][idx][:]
                    data_i = self._lght_to_grid(lght_data, t_slice)
                else:
                    data_i = self._hdf_files[fname][data_type][idx:idx + 1, :, :, t_slice]
                # Downsample t
                t_slice = [slice(None, None), ] * 4
                t_slice[-1] = slice(None, None, downsample_dict[data_type][0])  # layout = 'NHWT'
                data_i = data_i[tuple(t_slice)]
                # Downsample h, w
                data_i = block_reduce(data_i,
                                      block_size=(1, *downsample_dict[data_type][1:], 1),
                                      func=np.max)
                # Save as new .h5 file
                new_file_path = os.path.join(save_dir, fname)
                if not os.path.exists(new_file_path):
                    if not os.path.exists(os.path.dirname(new_file_path)):
                        os.makedirs(os.path.dirname(new_file_path))
                    # Create dataset
                    with h5py.File(new_file_path, 'w') as hf:
                        hf.create_dataset(
                            data_type, data=data_i,
                            maxshape=(None, *data_i.shape[1:]))
                else:
                    # Append
                    with h5py.File(new_file_path, 'a') as hf:
                        hf[data_type].resize((hf[data_type].shape[0] + data_i.shape[0]), axis=0)
                        hf[data_type][-data_i.shape[0]:] = data_i

            sample_counter += 1

    def save_downsampled_dataset(self, save_dir, downsample_dict, verbose=True):
        """
        Parameters
        ----------
        save_dir
        downsample_dict:    Dict[Sequence[int]]
            Notice that this is different from `self.downsample_dict`, which is used during runtime.
        """
        import os
        from skimage.measure import block_reduce
        from ...utils.utils import path_splitall
        assert not os.path.exists(save_dir), f"save_dir {save_dir} already exists!"
        os.makedirs(save_dir)
        for fname, hdf_file in self._hdf_files.items():
            if verbose:
                print(f"Downsampling data in {fname}.")
            data_type = path_splitall(fname)[0]
            if data_type == 'lght':
                # TODO: how to get idx?
                raise NotImplementedError
                # lght_data = self._hdf_files[fname][idx][:]
                # t_slice = slice(0, None)
                # data_i = self._lght_to_grid(lght_data, t_slice)
            else:
                data_i = self._hdf_files[fname][data_type]
            # Downsample t
            t_slice = [slice(None, None), ] * 4
            t_slice[-1] = slice(None, None, downsample_dict[data_type][0])  # layout = 'NHWT'
            data_i = data_i[tuple(t_slice)]
            # Downsample h, w
            data_i = block_reduce(data_i,
                                  block_size=(1, *downsample_dict[data_type][1:], 1),
                                  func=np.max)
            # Save as new .h5 file
            new_file_path = os.path.join(save_dir, fname)
            if not os.path.exists(os.path.dirname(new_file_path)):
                os.makedirs(os.path.dirname(new_file_path))
            # Create dataset
            with h5py.File(new_file_path, 'w') as hf:
                hf.create_dataset(
                    data_type, data=data_i,
                    maxshape=(None, *data_i.shape[1:]))

    @property
    def sample_count(self):
        """
        Record how many times self.__next__() is called.
        """
        return self._sample_count

    def inc_sample_count(self):
        self._sample_count += 1

    @property
    def curr_event_idx(self):
        return self._curr_event_idx

    @property
    def curr_seq_idx(self):
        """
        Used only when self.sample_mode == 'sequent'
        """
        return self._curr_seq_idx

    def set_curr_event_idx(self, val):
        self._curr_event_idx = val

    def set_curr_seq_idx(self, val):
        """
        Used only when self.sample_mode == 'sequent'
        """
        self._curr_seq_idx = val

    def reset(self, shuffle: bool = None):
        self.set_curr_event_idx(val=self.start_event_idx)
        self.set_curr_seq_idx(0)
        self._sample_count = 0
        if shuffle is None:
            shuffle = self.shuffle
        if shuffle:
            self.shuffle_samples()

    def __len__(self):
        """
        Used only when self.sample_mode == 'sequent'
        """
        return self.total_num_seq // self.batch_size

    @property
    def use_up(self):
        """
        Check if dataset is used up in 'sequent' mode.
        """
        if self.sample_mode == 'random':
            return False
        else:   # self.sample_mode == 'sequent'
            # compute the remaining number of sequences in current event
            curr_event_remain_seq = self.num_seq_per_event - self.curr_seq_idx
            all_remain_seq = curr_event_remain_seq + (
                        self.end_event_idx - self.curr_event_idx - 1) * self.num_seq_per_event
            if self.split_mode == "floor":
                # This approach does not cover all available data, but avoid dealing with masks
                return all_remain_seq < self.batch_size
            else:
                return all_remain_seq <= 0

    def _load_event_batch(self, event_idx, event_batch_size):
        """
        Loads a selected batch of events (not batch of sequences) into memory.

        Parameters
        ----------
        idx
        event_batch_size
            event_batch[i] = all_type_i_available_events[idx:idx + event_batch_size]
        Returns
        -------
        event_batch
            list of event batches.
            event_batch[i] is the event batch of the i-th data type.
            Each event_batch[i] is a np.ndarray with shape = (event_batch_size, height, width, raw_seq_len)
        """
        event_idx_slice_end = event_idx + event_batch_size
        pad_size = 0
        if event_idx_slice_end > self.end_event_idx:
            pad_size = event_idx_slice_end - self.end_event_idx
            event_idx_slice_end = self.end_event_idx
        pd_batch = self._samples.iloc[event_idx:event_idx_slice_end]
        data = {}
        for index, row in pd_batch.iterrows():
            data = self._read_data(row, data)
        if pad_size > 0:
            event_batch = []
            for t in self.data_types:
                pad_shape = [pad_size, ] + list(data[t].shape[1:])
                data_pad = np.concatenate((data[t].astype(self.output_type),
                                           np.zeros(pad_shape, dtype=self.output_type)),
                                          axis=0)
                event_batch.append(data_pad)
        else:
            event_batch = [data[t].astype(self.output_type) for t in self.data_types]
        return event_batch

    def __iter__(self):
        return self

    def __next__(self):
        if self.sample_mode == 'random':
            self.inc_sample_count()
            ret_dict = self._random_sample()
        else:
            if self.use_up:
                raise StopIteration
            else:
                self.inc_sample_count()
                ret_dict = self._sequent_sample()
        ret_dict = self.data_dict_to_tensor(data_dict=ret_dict,
                                            data_types=self.data_types)
        if self.preprocess:
            ret_dict = self.preprocess_data_dict(data_dict=ret_dict,
                                                 data_types=self.data_types,
                                                 layout=self.layout,
                                                 rescale=self.rescale_method)
        if self.downsample_dict is not None:
            ret_dict = self.downsample_data_dict(data_dict=ret_dict,
                                                 data_types=self.data_types,
                                                 factors_dict=self.downsample_dict,
                                                 layout=self.layout)
        return ret_dict

    def __getitem__(self, index):
        data_dict = self._idx_sample(index=index)
        return data_dict

    @staticmethod
    def preprocess_data_dict(data_dict, data_types=None, layout='NHWT', rescale='01'):
        """
        Parameters
        ----------
        data_dict:  Dict[str, Union[np.ndarray, torch.Tensor]]
        data_types: Sequence[str]
            The data types that we want to rescale. This mainly excludes "mask" from preprocessing.
        layout: str
            consists of batch_size 'N', seq_len 'T', channel 'C', height 'H', width 'W'
        rescale:    str
            'sevir': use the offsets and scale factors in original implementation.
            '01': scale all values to range 0 to 1, currently only supports 'vil'
        Returns
        -------
        data_dict:  Dict[str, Union[np.ndarray, torch.Tensor]]
            preprocessed data
        """
        if rescale == 'sevir':
            scale_dict = PREPROCESS_SCALE_SEVIR
            offset_dict = PREPROCESS_OFFSET_SEVIR
        elif rescale == '01':
            scale_dict = PREPROCESS_SCALE_01
            offset_dict = PREPROCESS_OFFSET_01
        else:
            raise ValueError(f'Invalid rescale option: {rescale}.')
        if data_types is None:
            data_types = data_dict.keys()
        for key, data in data_dict.items():
            if key in data_types:
                if isinstance(data, np.ndarray):
                    data = scale_dict[key] * (
                            data.astype(np.float32) +
                            offset_dict[key])
                    data = change_layout_np(data=data,
                                            in_layout='NHWT',
                                            out_layout=layout)
                elif isinstance(data, torch.Tensor):
                    data = scale_dict[key] * (
                            data.float() +
                            offset_dict[key])
                    data = change_layout_torch(data=data,
                                               in_layout='NHWT',
                                               out_layout=layout)
                data_dict[key] = data
        return data_dict

    @staticmethod
    def process_data_dict_back(data_dict, data_types=None, rescale='01'):
        """
        Parameters
        ----------
        data_dict
            each data_dict[key] is a torch.Tensor.
        rescale
            str:
                'sevir': data are scaled using the offsets and scale factors in original implementation.
                '01': data are all scaled to range 0 to 1, currently only supports 'vil'
        Returns
        -------
        data_dict
            each data_dict[key] is the data processed back in torch.Tensor.
        """
        if rescale == 'sevir':
            scale_dict = PREPROCESS_SCALE_SEVIR
            offset_dict = PREPROCESS_OFFSET_SEVIR
        elif rescale == '01':
            scale_dict = PREPROCESS_SCALE_01
            offset_dict = PREPROCESS_OFFSET_01
        else:
            raise ValueError(f'Invalid rescale option: {rescale}.')
        if data_types is None:
            data_types = data_dict.keys()
        for key in data_types:
            data = data_dict[key]
            data = data.float() / scale_dict[key] - offset_dict[key]
            data_dict[key] = data
        return data_dict

    @staticmethod
    def data_dict_to_tensor(data_dict, data_types=None):
        """
        Convert each element in data_dict to torch.Tensor (copy without grad).
        """
        ret_dict = {}
        if data_types is None:
            data_types = data_dict.keys()
        for key, data in data_dict.items():
            if key in data_types:
                if isinstance(data, torch.Tensor):
                    ret_dict[key] = data.detach().clone()
                elif isinstance(data, np.ndarray):
                    ret_dict[key] = torch.from_numpy(data)
                else:
                    raise ValueError(f"Invalid data type: {type(data)}. Should be torch.Tensor or np.ndarray")
            else:   # key == "mask"
                ret_dict[key] = data
        return ret_dict

    @staticmethod
    def downsample_data_dict(data_dict, data_types=None, factors_dict=None, layout='NHWT'):
        """
        Parameters
        ----------
        data_dict:  Dict[str, Union[np.array, torch.Tensor]]
        factors_dict:   Optional[Dict[str, Sequence[int]]]
            each element `factors` is a Sequence of int, representing (t_factor, h_factor, w_factor)

        Returns
        -------
        downsampled_data_dict:  Dict[str, torch.Tensor]
            Modify on a deep copy of data_dict instead of directly modifying the original data_dict
        """
        if factors_dict is None:
            factors_dict = {}
        if data_types is None:
            data_types = data_dict.keys()
        downsampled_data_dict = SEVIRDataLoader.data_dict_to_tensor(
            data_dict=data_dict,
            data_types=data_types)    # make a copy
        for key, data in data_dict.items():
            factors = factors_dict.get(key, None)
            if factors is not None:
                downsampled_data_dict[key] = change_layout_torch(
                    data=downsampled_data_dict[key],
                    in_layout=layout,
                    out_layout='NTHW')
                # downsample t dimension
                t_slice = [slice(None, None), ] * 4
                t_slice[1] = slice(None, None, factors[0])
                downsampled_data_dict[key] = downsampled_data_dict[key][tuple(t_slice)]
                # downsample spatial dimensions
                downsampled_data_dict[key] = avg_pool2d(
                    input=downsampled_data_dict[key],
                    kernel_size=(factors[1], factors[2]))

                downsampled_data_dict[key] = change_layout_torch(
                    data=downsampled_data_dict[key],
                    in_layout='NTHW',
                    out_layout=layout)

        return downsampled_data_dict

    def _random_sample(self):
        """
        Returns
        -------
        ret_dict
            dict. ret_dict.keys() == self.data_types.
            If self.preprocess == False:
                ret_dict[imgt].shape == (batch_size, height, width, seq_len)
        """
        num_sampled = 0
        event_idx_list = nprand.randint(low=self.start_event_idx,
                                        high=self.end_event_idx,
                                        size=self.batch_size)
        seq_idx_list = nprand.randint(low=0,
                                      high=self.num_seq_per_event,
                                      size=self.batch_size)
        seq_slice_list = [slice(seq_idx * self.stride,
                                seq_idx * self.stride + self.seq_len)
                          for seq_idx in seq_idx_list]
        ret_dict = {}
        while num_sampled < self.batch_size:
            event = self._load_event_batch(event_idx=event_idx_list[num_sampled],
                                           event_batch_size=1)
            for imgt_idx, imgt in enumerate(self.data_types):
                sampled_seq = event[imgt_idx][[0, ], :, :, seq_slice_list[num_sampled]]  # keep the dim of batch_size for concatenation
                if imgt in ret_dict:
                    ret_dict[imgt] = np.concatenate((ret_dict[imgt], sampled_seq),
                                                    axis=0)
                else:
                    ret_dict.update({imgt: sampled_seq})
        return ret_dict

    def _sequent_sample(self):
        """
        Returns
        -------
        ret_dict:   Dict
            `ret_dict.keys()` contains `self.data_types`.
            `ret_dict["mask"]` is a list of bool, indicating if the data entry is real or padded.
            If self.preprocess == False:
                ret_dict[imgt].shape == (batch_size, height, width, seq_len)
        """
        assert not self.use_up, 'Data loader used up! Reset it to reuse.'
        event_idx = self.curr_event_idx
        seq_idx = self.curr_seq_idx
        num_sampled = 0
        sampled_idx_list = []   # list of (event_idx, seq_idx) records
        while num_sampled < self.batch_size:
            sampled_idx_list.append({'event_idx': event_idx,
                                     'seq_idx': seq_idx})
            seq_idx += 1
            if seq_idx >= self.num_seq_per_event:
                event_idx += 1
                seq_idx = 0
            num_sampled += 1

        start_event_idx = sampled_idx_list[0]['event_idx']
        event_batch_size = sampled_idx_list[-1]['event_idx'] - start_event_idx + 1

        event_batch = self._load_event_batch(event_idx=start_event_idx,
                                             event_batch_size=event_batch_size)
        ret_dict = {"mask": []}
        all_no_pad_flag = True
        for sampled_idx in sampled_idx_list:
            batch_slice = [sampled_idx['event_idx'] - start_event_idx, ]  # use [] to keepdim
            seq_slice = slice(sampled_idx['seq_idx'] * self.stride,
                              sampled_idx['seq_idx'] * self.stride + self.seq_len)
            for imgt_idx, imgt in enumerate(self.data_types):
                sampled_seq = event_batch[imgt_idx][batch_slice, :, :, seq_slice]
                if imgt in ret_dict:
                    ret_dict[imgt] = np.concatenate((ret_dict[imgt], sampled_seq),
                                                    axis=0)
                else:
                    ret_dict.update({imgt: sampled_seq})
            # add mask
            no_pad_flag = sampled_idx['event_idx'] < self.end_event_idx
            if not no_pad_flag:
                all_no_pad_flag = False
            ret_dict["mask"].append(no_pad_flag)
        if all_no_pad_flag:
            # if there is no padded data items at all, set `ret_dict["mask"] = None` for convenience.
            ret_dict["mask"] = None
        # update current idx
        self.set_curr_event_idx(event_idx)
        self.set_curr_seq_idx(seq_idx)
        return ret_dict

    def _idx_sample(self, index):
        """
        Parameters
        ----------
        index
            The index of the batch to sample.
        Returns
        -------
        ret_dict
            dict. ret_dict.keys() == self.data_types.
            If self.preprocess == False:
                ret_dict[imgt].shape == (batch_size, height, width, seq_len)
        """
        event_idx = (index * self.batch_size) // self.num_seq_per_event
        seq_idx = (index * self.batch_size) % self.num_seq_per_event
        num_sampled = 0
        sampled_idx_list = []  # list of (event_idx, seq_idx) records
        while num_sampled < self.batch_size:
            sampled_idx_list.append({'event_idx': event_idx,
                                     'seq_idx': seq_idx})
            seq_idx += 1
            if seq_idx >= self.num_seq_per_event:
                event_idx += 1
                seq_idx = 0
            num_sampled += 1

        start_event_idx = sampled_idx_list[0]['event_idx']
        event_batch_size = sampled_idx_list[-1]['event_idx'] - start_event_idx + 1

        event_batch = self._load_event_batch(event_idx=start_event_idx,
                                             event_batch_size=event_batch_size)
        ret_dict = {}
        for sampled_idx in sampled_idx_list:
            batch_slice = [sampled_idx['event_idx'] - start_event_idx, ]  # use [] to keepdim
            seq_slice = slice(sampled_idx['seq_idx'] * self.stride,
                              sampled_idx['seq_idx'] * self.stride + self.seq_len)
            for imgt_idx, imgt in enumerate(self.data_types):
                sampled_seq = event_batch[imgt_idx][batch_slice, :, :, seq_slice]
                if imgt in ret_dict:
                    ret_dict[imgt] = np.concatenate((ret_dict[imgt], sampled_seq),
                                                    axis=0)
                else:
                    ret_dict.update({imgt: sampled_seq})

        ret_dict = self.data_dict_to_tensor(data_dict=ret_dict,
                                            data_types=self.data_types)
        if self.preprocess:
            ret_dict = self.preprocess_data_dict(data_dict=ret_dict,
                                                 data_types=self.data_types,
                                                 layout=self.layout,
                                                 rescale=self.rescale_method)

        if self.downsample_dict is not None:
            ret_dict = self.downsample_data_dict(data_dict=ret_dict,
                                                 data_types=self.data_types,
                                                 factors_dict=self.downsample_dict,
                                                 layout=self.layout)
        return ret_dict



from typing import Sequence
import random
import torchvision.transforms.functional as TF


class TransformsFixRotation:
    r"""
    Rotate by one of the given angles.

    Example: `rotation_transform = MyRotationTransform(angles=[-30, -15, 0, 15, 30])`
    """

    def __init__(self, angles):
        if not isinstance(angles, Sequence):
            angles = [angles, ]
        self.angles = angles

    def __call__(self, x):
        angle = random.choice(self.angles)
        return TF.rotate(x, angle)





import xarray as xr

import torch
import os
from train_cuboid_earthnet import CuboidEarthNet2021PLModule
from earthformer.utils.utils import download
from omegaconf import OmegaConf
from earthformer.datasets.earthnet.earthnet21x_dataloader import EarthNet2021xLightningDataModule, EarthNet2021xTestDataset
from pytorch_lightning import Trainer, seed_everything, loggers as pl_loggers
from earthformer.datasets.earthnet.visualization import vis_earthnet_seq
import numpy as np
from pathlib import Path
from tqdm import tqdm


def process_sample(sample):
    # High-resolution Earth surface data. The channels are [blue, green, red, nir, cloud]
    highresdynamic = sample['highresdynamic']
    highresstatic = sample['highresstatic']

    # The meso-scale data. The channels are ["precipitation", "pressure", "temp mean", "temp min", "temp max"]
    mesodynamic = sample['mesodynamic']
    mesostatic = sample['mesostatic']

    highresdynamic = np.nan_to_num(highresdynamic, nan=0.0, posinf=1.0, neginf=0.0)
    highresdynamic = np.clip(highresdynamic, a_min=0.0, a_max=1.0)
    mesodynamic = np.nan_to_num(mesodynamic, nan=0.0)
    highresstatic = np.nan_to_num(highresstatic, nan=0.0)
    mesostatic = np.nan_to_num(mesostatic, nan=0.0)
    return highresdynamic, highresstatic, mesodynamic, mesostatic

def main():
    print("Loading Config")
    pred_dir = Path("./experiments/preds_en21x/")

    config_file = "./earthformer_earthnet_v1.yaml"
    config = OmegaConf.load(open(config_file, "r"))
    in_len = config.layout.in_len
    out_len = config.layout.out_len

    seed = config.optim.seed
    dataset_cfg = OmegaConf.to_object(config.dataset)

    seed_everything(seed, workers=True)

    micro_batch_size = 1
    print("Loading Dataset")
    earthnet_iid_testset = EarthNet2021xTestDataset(subset_name="iid",
                                                data_dir="/Net/Groups/BGI/work_1/scratch/s3/earthnet/earthnet2021x/iid/",
                                                layout=config.dataset.layout,
                                                static_layout=config.dataset.static_layout,
                                                highresstatic_expand_t=config.dataset.highresstatic_expand_t,
                                                mesostatic_expand_t=config.dataset.mesostatic_expand_t,
                                                meso_crop=None,
                                                fp16=False)

    save_dir = "./experiments"
    print("Loading Model")
    pl_module = CuboidEarthNet2021PLModule(
        total_num_steps=None,
        save_dir="./experiments",
        oc_file=config_file
    )

    pretrained_checkpoint_url = "https://earthformer.s3.amazonaws.com/pretrained_checkpoints/earthformer_earthnet2021.pt"
    local_checkpoint_path = os.path.join(save_dir, "earthformer_earthnet2021.pt")
    download(url=pretrained_checkpoint_url, path=local_checkpoint_path)

    state_dict = torch.load(local_checkpoint_path, map_location=torch.device("cpu"))
    pl_module.torch_nn_module.load_state_dict(state_dict=state_dict)

    pl_module.torch_nn_module.cuda()
    pl_module.torch_nn_module.eval()
    print("Starting Predictions")
    for idx in tqdm(range(len(earthnet_iid_testset))):
        highresdynamic, highresstatic, mesodynamic, mesostatic = process_sample(earthnet_iid_testset[idx])


        with torch.no_grad():
            pred_seq, loss, in_seq, target_seq, mask = pl_module({"highresdynamic": torch.tensor(np.expand_dims(highresdynamic, axis=0)).cuda(), 
                                                                "highresstatic": torch.tensor(np.expand_dims(highresstatic, axis=0)).cuda(),
                                                                "mesodynamic": torch.tensor(np.expand_dims(mesodynamic, axis=0)).cuda(),
                                                                "mesostatic": torch.tensor(np.expand_dims(mesostatic, axis=0)).cuda()})
            pred_seq_np = pred_seq.detach().cpu().numpy()

        targ_path = Path(earthnet_iid_testset.nc_path_list[idx])

        targ_cube = xr.open_dataset(targ_path)

        lat = targ_cube.lat
        lon = targ_cube.lon

        blue_pred = pred_seq_np[0,:,:,:,0]
        green_pred = pred_seq_np[0,:,:,:,1]
        red_pred = pred_seq_np[0,:,:,:,2]
        nir_pred = pred_seq_np[0,:,:,:,3]
        ndvi_pred = ((nir_pred - red_pred) / (nir_pred + red_pred + 1e-8))
        
        pred_cube = xr.Dataset({"ndvi_pred": xr.DataArray(data = ndvi_pred, coords = {"time": targ_cube.time.isel(time = slice(4,None,5)).isel(time = slice(in_len, in_len + out_len)), "lat": lat, "lon": lon}, dims = ["time","lat", "lon"])})

        
        pred_path = pred_dir/targ_path.parent.stem/targ_path.name
        pred_path.parent.mkdir(parents = True, exist_ok = True)
        if not pred_path.is_file():
            pred_cube.to_netcdf(pred_path, encoding={"ndvi_pred":{"dtype": "float32"}})

if __name__ == "__main__":
    main()


import torch
from torch import nn
from torch.nn import functional as F


class Persistence(nn.Module):

    def __init__(self, layout="NTHWC"):
        super(Persistence, self).__init__()
        self.layout = layout
        self.t_axis = self.layout.find("T")
        self.in_seq_slice = [slice(None, None), ] * len(self.layout)
        self.in_seq_slice[self.t_axis] = slice(-1, None)

    def forward(self, in_seq, out_seq):
        out_len = out_seq.shape[self.t_axis]
        output = in_seq[self.in_seq_slice]
        output = torch.repeat_interleave(output,
                                         repeats=out_len,
                                         dim=self.t_axis)
        loss = F.mse_loss(output, out_seq)
        return output, loss



import os
from omegaconf import OmegaConf

_CURR_DIR = os.path.realpath(os.path.dirname(os.path.realpath(__file__)))


cfg = OmegaConf.create()
cfg.root_dir = os.path.abspath(os.path.realpath(os.path.join(_CURR_DIR, "..", "..")))
cfg.datasets_dir = os.path.join(cfg.root_dir, "datasets")  # default directory for loading datasets
cfg.pretrained_checkpoints_dir = os.path.join(cfg.root_dir, "pretrained_checkpoints")  # default directory for saving and loading pretrained checkpoints
cfg.exps_dir = os.path.join(cfg.root_dir, "experiments")  # default directory for saving experiment results
os.makedirs(cfg.exps_dir, exist_ok=True)



from typing import Optional, Union, Sequence, Dict
import os
from pathlib import Path
import numpy as np
import torch
from torch import nn
from torch.utils.data import Dataset, Subset, DataLoader, random_split
import torchvision.transforms as T
from pytorch_lightning import LightningDataModule
from einops import rearrange
from ..augmentation import TransformsFixRotation
from ...config import cfg


default_data_dir = os.path.join(cfg.datasets_dir, "earthnet2021")

def change_layout(data: np.ndarray,
                  in_layout: str = "HWCT",
                  out_layout: str = "HWCT"):
    axes = [None, ] * len(in_layout)
    for i, axis in enumerate(in_layout):
        axes[out_layout.find(axis)] = i
    return np.transpose(data, axes=axes)

def einops_change_layout(data: Union[np.ndarray, torch.Tensor],
                         einops_in_layout: str = "H W C T",
                         einops_out_layout: str = "H W C T"):
    return rearrange(data, f"{einops_in_layout} -> {einops_out_layout}")

class _BaseEarthNet2021Dataset(Dataset):
    r"""
    An .npy file contains a dict with

    "highresdynamic":   np.ndarray
        shape = (128, 128, 7, T_highres)
        channels are [blue, green, red, nir, cloud, scene, mask]
    "highresstatic":    np.ndarray
        shape = (128, 128, 1)
        channel is [elevation]
    "mesodynamic":      np.ndarray
        shape = (80, 80, 5, T_meso)
        channels are [precipitation, pressure, temp mean, temp min, temp max]
    "mesostatic":       np.ndarray
        shape = (80, 80, 1)
        channel is [elevation]

    train:
        T_highres = 30, T_meso = 150
    iid/ood test:
        T_highres = 10, T_meso = 150 for context
        T_highres = 20 for target
    extreme test:
        T_highres = 20, T_meso = 300 for context
        T_highres = 40 for target
    seasonal test:
        T_highres = 70, T_meso = 1050 for context
        T_highres = 140 for target
    """
    default_layout = "HWCT"
    default_static_layout = "HWC"
    # flip requires the last two dims to be H,W
    layout_for_aug = "CTHW"
    static_layout_for_aug = "CHW"

    def __init__(self,
                 return_mode: str = "default",
                 data_aug_mode: str = None,
                 data_aug_cfg: Dict = None,
                 layout: str = default_layout,
                 static_layout: str = default_static_layout,
                 highresstatic_expand_t: bool = False,
                 mesostatic_expand_t: bool = False,
                 meso_crop: Union[str, Sequence[Sequence[int]]] = None,
                 fp16: bool = False, ):
        r"""

        Parameters
        ----------
        return_mode:    str
            "default":
                return {
                    "highresdynamic": highresdynamic,
                    "highresstatic": highresstatic,
                    "mesodynamic": mesodynamic,
                    "mesostatic": mesostatic,
                }
            "minimal":
                return highresdynamic[..., 4, :], i.e., only RGB and IR channels.
        data_aug_mode:  str
            If None, no data augmentation is performed
            If "0", apply `RandomHorizontalFlip(p=0.5)` and RandomVerticalFlip(p=0.5)
        layout: str
            The layout of returned dynamic data ndarray.
        static_layout:  str
            The layout of returned static data ndarray. Take no effect if expanding temporal dim.
        highresstatic_expand_t: bool
            If True, add a new temporal dim for highresstatic data, use the same layout as dynamic data.
        mesostatic_expand_t:    bool
            If True, add a new temporal dim for mesostatic data, use the same layout as dynamic data.
        meso_crop:  Union[str, Sequence[Sequence[int]]]
            If None, take no effect
            If "default", use `((39, 41), (39, 41))` to crop out overlapping section with highres
            Can also be specified arbitrarily in form `((H_s, H_e), (W_s, W_e))`.
        fp16:   bool
            Use np.float16 if True else np.float32
        """
        self.return_mode = return_mode
        self.data_aug_mode = data_aug_mode
        if data_aug_cfg is None:
            data_aug_cfg = {}
        self.data_aug_cfg = data_aug_cfg
        if self.data_aug_mode is None:
            pass
        elif self.data_aug_mode in ["0", ]:
            self.data_aug = T.Compose([
                T.RandomHorizontalFlip(p=0.5),
                T.RandomVerticalFlip(p=0.5),
            ])
        elif self.data_aug_mode in ["1", "2"]:
            self.data_aug = T.Compose([
                T.RandomHorizontalFlip(p=0.5),
                T.RandomVerticalFlip(p=0.5),
                TransformsFixRotation([0, 90, 180, 270]),
            ])
        else:
            raise NotImplementedError
        if layout is None:
            layout = self.default_layout
        self.layout = layout
        if static_layout is None:
            static_layout = self.default_static_layout
        self.static_layout = static_layout
        self.highresstatic_expand_t = highresstatic_expand_t
        self.mesostatic_expand_t = mesostatic_expand_t
        if isinstance(meso_crop, str):
            assert meso_crop == "default", f"meso_crop mode {meso_crop} not supported."
            meso_crop = ((39, 41), (39, 41))
        self.meso_crop = meso_crop
        self.np_dtype = np.float16 if fp16 else np.float32

    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, idx):
        raise NotImplementedError

    @property
    def einops_default_layout(self):
        if not hasattr(self, "_einops_default_layout"):
            self._einops_default_layout = " ".join(self.default_layout)
        return self._einops_default_layout

    @property
    def einops_default_static_layout(self):
        if not hasattr(self, "_einops_default_static_layout"):
            self._einops_default_static_layout = " ".join(self.default_static_layout)
        return self._einops_default_static_layout

    @property
    def einops_layout(self):
        if not hasattr(self, "_einops_layout"):
            self._einops_layout = " ".join(self.layout)
        return self._einops_layout

    @property
    def einops_static_layout(self):
        if not hasattr(self, "_einops_static_layout"):
            self._einops_static_layout = " ".join(self.static_layout)
        return self._einops_static_layout

    @property
    def einops_layout_for_aug(self):
        if not hasattr(self, "_einops_layout_for_aug"):
            self._einops_layout_for_aug = " ".join(self.layout_for_aug)
        return self._einops_layout_for_aug

    @property
    def einops_static_layout_for_aug(self):
        if not hasattr(self, "_einops_static_layout_for_aug"):
            self._einops_static_layout_for_aug = " ".join(self.static_layout_for_aug)
        return self._einops_static_layout_for_aug

    def process_raw_data_from_npz(
            self,
            highresdynamic, highresstatic,
            mesodynamic, mesostatic):
        r"""
        Process np.ndarray data loaded from saved .npz files

        Parameters
        ----------
        highresdynamic:   np.ndarray
            shape = (128, 128, 7, T_highres)
            channels are [blue, green, red, nir, cloud, scene, mask]
        highresstatic:    np.ndarray
            shape = (128, 128, 1)
            channel is [elevation]
        mesodynamic:      np.ndarray
            shape = (80, 80, 5, T_meso)
            channels are [precipitation, pressure, temp mean, temp min, temp max]
        mesostatic:       np.ndarray
            shape = (80, 80, 1)
            channel is [elevation]

            (T_highres, T_meso) =
                (30, 150) for train, iid, ood
                (60, 300) for extreme
                (210, 1050) for seasonal

        Returns
        -------
        ret:    dict
        """
        T_highres = highresdynamic.shape[-1]
        T_meso = mesodynamic.shape[-1]

        if self.return_mode in ["default", ]:
            if self.meso_crop is not None:
                mesodynamic = self.crop_meso_spatial(mesodynamic)
                mesostatic = self.crop_meso_spatial(mesostatic)

            highresdynamic = np.nan_to_num(highresdynamic, copy=False, nan=0.0, posinf=1.0, neginf=0.0)
            highresdynamic = np.clip(highresdynamic, a_min=0.0, a_max=1.0)
            mesodynamic = np.nan_to_num(mesodynamic, copy=False, nan=0.0)
            highresstatic = np.nan_to_num(highresstatic, copy=False, nan=0.0)
            mesostatic = np.nan_to_num(mesostatic, copy=False, nan=0.0)

            highresdynamic = einops_change_layout(
                data=highresdynamic,
                einops_in_layout=self.einops_default_layout,
                einops_out_layout=self.einops_layout)
            mesodynamic = einops_change_layout(
                data=mesodynamic,
                einops_in_layout=self.einops_default_layout,
                einops_out_layout=self.einops_layout)
            if self.highresstatic_expand_t:
                highresstatic = np.repeat(highresstatic[..., np.newaxis],
                                          repeats=T_highres,
                                          axis=-1)
                highresstatic = einops_change_layout(
                    data=highresstatic,
                    einops_in_layout=self.einops_default_layout,
                    einops_out_layout=self.einops_layout)
            else:
                highresstatic = einops_change_layout(
                    data=highresstatic,
                    einops_in_layout=self.einops_default_static_layout,
                    einops_out_layout=self.einops_static_layout)
            if self.mesostatic_expand_t:
                mesostatic = np.repeat(mesostatic[..., np.newaxis],
                                       repeats=T_meso,
                                       axis=-1)
                mesostatic = einops_change_layout(
                    data=mesostatic,
                    einops_in_layout=self.einops_default_layout,
                    einops_out_layout=self.einops_layout)
            else:
                mesostatic = einops_change_layout(
                    data=mesostatic,
                    einops_in_layout=self.einops_default_static_layout,
                    einops_out_layout=self.einops_static_layout)
            if self.return_mode == "default":
                if self.data_aug_mode is not None:
                    # TODO: augment all components consistently
                    raise NotImplementedError
                return {
                    "highresdynamic": highresdynamic,
                    "highresstatic": highresstatic,
                    "mesodynamic": mesodynamic,
                    "mesostatic": mesostatic,
                }
            else:
                raise NotImplementedError
        elif self.return_mode in ["minimal", ]:
            # only RGB, infrared channels and mask
            highresdynamic = np.nan_to_num(highresdynamic, copy=False, nan=0.0, posinf=1.0, neginf=0.0)
            highresdynamic = np.clip(highresdynamic, a_min=0.0, a_max=1.0)

            if self.data_aug_mode is not None:
                highresdynamic = einops_change_layout(
                    data=highresdynamic,
                    einops_in_layout=self.einops_default_layout,
                    einops_out_layout=self.einops_layout_for_aug)
                highresdynamic = self.data_aug(torch.from_numpy(highresdynamic))
                highresdynamic = einops_change_layout(
                    data=highresdynamic,
                    einops_in_layout=self.einops_layout_for_aug,
                    einops_out_layout=self.einops_layout).numpy()
            else:
                highresdynamic = einops_change_layout(
                    data=highresdynamic,
                    einops_in_layout=self.einops_default_layout,
                    einops_out_layout=self.einops_layout)
            return highresdynamic
        else:
            raise NotImplementedError(f"return_mode {self.return_mode} not supported!")

    def crop_meso_spatial(self, meso_data):
        r"""
        Crop the meso data along spatial dims, under default layout.
        """
        if self.meso_crop is None:
            return meso_data
        else:
            return meso_data[
                   self.meso_crop[0][0]:self.meso_crop[0][1],
                   self.meso_crop[1][0]:self.meso_crop[1][1],
                   ...]

class EarthNet2021TrainDataset(_BaseEarthNet2021Dataset):
    r"""
    An .npy file contains a dict with
    "highresdynamic":   np.ndarray
        shape = (128, 128, 7, 30)
        channels are [blue, green, red, nir, cloud, scene, mask]
    "highresstatic":    np.ndarray
        shape = (128, 128, 1)
        channel is [elevation]
    "mesodynamic":      np.ndarray
        shape = (80, 80, 5, 150)
        channels are [precipitation, pressure, temp mean, temp min, temp max]
    "mesostatic":       np.ndarray
        shape = (80, 80, 1)
        channel is [elevation]
    """
    default_train_dir = os.path.join(default_data_dir, "train")

    T_highres = 30
    T_meso = 150

    def __init__(self,
                 return_mode: str = "default",
                 data_aug_mode: str = None,
                 data_aug_cfg: Dict = None,
                 data_dir: Union[Path, str] = None,
                 layout: str = None,
                 static_layout: str = None,
                 highresstatic_expand_t: bool = False,
                 mesostatic_expand_t: bool = False,
                 meso_crop: Union[str, Sequence[Sequence[int]]] = None,
                 fp16: bool = False):
        r"""

        Parameters
        ----------
        return_mode:    str
            "default":
                return {
                    "highresdynamic": highresdynamic,
                    "highresstatic": highresstatic,
                    "mesodynamic": mesodynamic,
                    "mesostatic": mesostatic,
                }
            "minimal":
                return highresdynamic[..., 4, :], i.e., only RGB and IR channels.
        data_aug_mode:  str
            If None, no data augmentation is performed
            If "0", apply `RandomHorizontalFlip(p=0.5)` and RandomVerticalFlip(p=0.5)
        data_aug_cfg:   dict
            dict which contains cfgs for controlling data augmentation
        data_dir:   Union[Path, str]
            Save dir of training data.
        layout: str
            The layout of returned dynamic data ndarray.
        static_layout:  str
            The layout of returned static data ndarray. Take no effect if expanding temporal dim.
        highresstatic_expand_t: bool
            If True, add a new temporal dim for highresstatic data, use the same layout as dynamic data.
        mesostatic_expand_t:    bool
            If True, add a new temporal dim for mesostatic data, use the same layout as dynamic data.
        meso_crop:  Union[str, Sequence[Sequence[int]]]
            If None, take no effect
            If "default", use `((39, 41), (39, 41))` to crop out overlapping section with highres
            Can also be specified arbitrarily in form `((H_s, H_e), (W_s, W_e))`.
        fp16:   bool
            Use np.float16 if True else np.float32
        """
        super(EarthNet2021TrainDataset, self).__init__(
            return_mode=return_mode,
            data_aug_mode=data_aug_mode,
            data_aug_cfg=data_aug_cfg,
            layout=layout,
            static_layout=static_layout,
            highresstatic_expand_t=highresstatic_expand_t,
            mesostatic_expand_t=mesostatic_expand_t,
            meso_crop=meso_crop,
            fp16=fp16, )
        if data_dir is None:
            data_dir = self.default_train_dir
        self.data_dir = Path(data_dir)
        self.npz_path_list = sorted(list(self.data_dir.glob("**/*.npz")))

    def __len__(self) -> int:
        return len(self.npz_path_list)

    def __getitem__(self, idx: int) -> dict:
        data_npz = np.load(self.npz_path_list[idx])

        # keep only [blue, green, red, nir, mask] channels
        highresdynamic = data_npz["highresdynamic"].astype(self.np_dtype)[:, :, [0, 1, 2, 3, 6], :]
        highresstatic = data_npz["highresstatic"].astype(self.np_dtype)
        mesodynamic = data_npz["mesodynamic"].astype(self.np_dtype)
        mesostatic = data_npz["mesostatic"].astype(self.np_dtype)

        if self.data_aug_mode in ["2", ]:
            processed_0 = self.process_raw_data_from_npz(
                highresdynamic, highresstatic,
                mesodynamic, mesostatic)
            data_npz_1 = np.load(self.npz_path_list[np.random.randint(len(self))])
            # keep only [blue, green, red, nir, mask] channels
            highresdynamic_1 = data_npz_1["highresdynamic"].astype(self.np_dtype)[:, :, [0, 1, 2, 3, 6], :]
            highresstatic_1 = data_npz_1["highresstatic"].astype(self.np_dtype)
            mesodynamic_1 = data_npz_1["mesodynamic"].astype(self.np_dtype)
            mesostatic_1 = data_npz_1["mesostatic"].astype(self.np_dtype)
            processed_1 = self.process_raw_data_from_npz(
                highresdynamic_1, highresstatic_1,
                mesodynamic_1, mesostatic_1)
            alpha = self.data_aug_cfg.get("mixup_alpha", 0.2)
            lam = np.random.beta(a=alpha, b=alpha)
            if self.return_mode in ["default", ]:
                ret = {}
                for key, val in processed_0.items():
                    ret[key] = lam * val + (1 - lam) * processed_1[key]
            elif self.return_mode in ["minimal", ]:
                ret = lam * processed_0 + (1 - lam) * processed_1
            else:
                raise NotImplementedError
            return ret
        else:
            return self.process_raw_data_from_npz(
                highresdynamic, highresstatic,
                mesodynamic, mesostatic)

class EarthNet2021TestDataset(_BaseEarthNet2021Dataset):
    r"""
    An .npy file contains a dict with
    "highresdynamic":   np.ndarray
        shape = (128, 128, 5, T_highres_context) for context
        shape = (128, 128, 5, T_highres_target) for target
        channels are [blue, green, red, nir, cloud, scene, mask]
    "highresstatic":    np.ndarray
        shape = (128, 128, 1)
        channel is [elevation]
    "mesodynamic":      np.ndarray
        shape = (80, 80, 5, T_meso) for context
    "mesostatic":       np.ndarray
        shape = (80, 80, 1)
        channel is [elevation]
    """
    default_iid_test_data_dir = os.path.join(default_data_dir, "iid_test_split")
    default_ood_test_data_dir = os.path.join(default_data_dir, "ood_test_split")
    default_extreme_test_data_dir = os.path.join(default_data_dir, "extreme_test_split")
    default_seasonal_test_data_dir = os.path.join(default_data_dir, "seasonal_test_split")

    T_highres_context = {"iid": 10, "ood": 10, "extreme": 20, "seasonal": 70}
    T_highres_target = {"iid": 20, "ood": 20, "extreme": 40, "seasonal": 140}
    T_highres = {key: val_context + val_target
                 for (key, val_context), (_, val_target) in
                 zip(T_highres_context.items(), T_highres_target.items())}
    T_meso = {"iid": 150, "ood": 150, "extreme": 300, "seasonal": 1050}

    def __init__(self,
                 return_mode: str = "default",
                 subset_name: str = "iid",
                 data_dir: Union[Path, str] = None,
                 layout: str = None,
                 static_layout: str = None,
                 highresstatic_expand_t: bool = False,
                 mesostatic_expand_t: bool = False,
                 meso_crop: Union[str, Sequence[Sequence[int]]] = None,
                 fp16: bool = False):
        r"""

        Parameters
        ----------
        return_mode:    str
            "default":
                return {
                    "highresdynamic": highresdynamic,
                    "highresstatic": highresstatic,
                    "mesodynamic": mesodynamic,
                    "mesostatic": mesostatic,
                }
            "minimal":
                return highresdynamic[..., 4, :], i.e., only RGB and IR channels.
        subset_name:    str
            Name of subset to load from default dir. Must be in ("iid", "ood", "extreme", "seasonal")
        data_dir:   Union[Path, str]
            if `subset_name` is None, use user specified dir
        layout: str
            The layout of returned dynamic data ndarray.
        static_layout:  str
            The layout of returned static data ndarray. Take no effect if expanding temporal dim.
        highresstatic_expand_t: bool
            If True, add a new temporal dim for highresstatic data, use the same layout as dynamic data.
        mesostatic_expand_t:    bool
            If True, add a new temporal dim for mesostatic data, use the same layout as dynamic data.
        meso_crop:  Union[str, Sequence[Sequence[int]]]
            If None, take no effect
            If "default", use `((39, 41), (39, 41))` to crop out overlapping section with highres
            Can also be specified arbitrarily in form `((H_s, H_e), (W_s, W_e))`.
        fp16:   bool
            Use np.float16 if True else np.float32
        """
        super(EarthNet2021TestDataset, self).__init__(
            return_mode=return_mode,
            data_aug_mode=None,
            layout=layout,
            static_layout=static_layout,
            highresstatic_expand_t=highresstatic_expand_t,
            mesostatic_expand_t=mesostatic_expand_t,
            meso_crop=meso_crop,
            fp16=fp16, )
        if subset_name == "iid":
            data_dir = self.default_iid_test_data_dir if data_dir is None else data_dir
        elif subset_name == "ood":
            data_dir = self.default_ood_test_data_dir if data_dir is None else data_dir
        elif subset_name == "extreme":
            data_dir = self.default_extreme_test_data_dir if data_dir is None else data_dir
        elif subset_name == "seasonal":
            data_dir = self.default_seasonal_test_data_dir if data_dir is None else data_dir
        else:
            assert subset_name is None  # Use user specified arg data_dir
        self.subset_name = subset_name
        self.data_dir = Path(data_dir)
        self.context_data_dir = self.data_dir.joinpath("context")
        self.target_data_dir = self.data_dir.joinpath("target")
        self.context_npz_path_list = sorted(list(self.context_data_dir.glob("**/*.npz")))
        self.target_npz_path_list = sorted(list(self.target_data_dir.glob("**/*.npz")))

    def __len__(self) -> int:
        return len(self.context_npz_path_list)

    def __getitem__(self, idx: int) -> dict:
        context_data_npz = np.load(self.context_npz_path_list[idx])
        target_data_npz = np.load(self.target_npz_path_list[idx])

        context_highresdynamic = context_data_npz["highresdynamic"].astype(self.np_dtype)
        highresstatic = context_data_npz["highresstatic"].astype(self.np_dtype)
        mesodynamic = context_data_npz["mesodynamic"].astype(self.np_dtype)
        mesostatic = context_data_npz["mesostatic"].astype(self.np_dtype)
        target_highresdynamic = target_data_npz["highresdynamic"].astype(self.np_dtype)
        highresdynamic = np.concatenate([context_highresdynamic,
                                         target_highresdynamic],
                                        axis=-1)

        return self.process_raw_data_from_npz(highresdynamic, highresstatic,
                                              mesodynamic, mesostatic)

class EarthNet2021LightningDataModule(LightningDataModule):

    def __init__(self,
                 return_mode: str = "default",
                 data_aug_mode: str = None,
                 data_aug_cfg: Dict = None,
                 train_data_dir: Union[Path, str] = None,
                 test_subset_name: Union[str, Sequence[str]] = ("iid", "ood"),
                 test_data_dir: Union[Union[Path, str], Sequence[Union[Path, str]]] = None,
                 val_ratio: float = 0.1,
                 train_val_split_seed: int = None,
                 layout: str = None,
                 static_layout: str = None,
                 highresstatic_expand_t: bool = False,
                 mesostatic_expand_t: bool = False,
                 meso_crop: Union[str, Sequence[Sequence[int]]] = None,
                 fp16: bool = False,
                 # datamodule_only
                 batch_size=1,
                 num_workers=8, ):
        super(EarthNet2021LightningDataModule, self).__init__()
        self.return_mode = return_mode
        self.data_aug_mode = data_aug_mode
        self.data_aug_cfg = data_aug_cfg
        if train_data_dir is None:
            train_data_dir = EarthNet2021TrainDataset.default_train_dir
        self.train_data_dir = train_data_dir

        if test_subset_name is None:
            if not isinstance(test_data_dir, Sequence):
                self.test_data_dir_list = [test_data_dir, ]
            else:
                self.test_data_dir_list = list(test_data_dir)
            self.test_subset_name_list = [None, ] * len(self.test_data_dir_list)
        else:
            if isinstance(test_subset_name, str):
                self.test_subset_name_list = [test_subset_name, ]
            elif isinstance(test_subset_name, Sequence):
                self.test_subset_name_list = list(test_subset_name)
            else:
                raise ValueError(f"Invalid type of test_subset_name {type(test_subset_name)}")
            self.test_data_dir_list = []
            for test_subset_name in self.test_subset_name_list:
                if test_subset_name == "iid":
                    test_data_dir = EarthNet2021TestDataset.default_iid_test_data_dir
                elif test_subset_name == "ood":
                    test_data_dir = EarthNet2021TestDataset.default_ood_test_data_dir
                elif test_subset_name == "extreme":
                    test_data_dir = EarthNet2021TestDataset.default_extreme_test_data_dir
                elif test_subset_name == "seasonal":
                    test_data_dir = EarthNet2021TestDataset.default_seasonal_test_data_dir
                else:
                    raise ValueError(f"Invalid test_subset_name {test_subset_name}")
                self.test_data_dir_list.append(test_data_dir)

        self.val_ratio = val_ratio
        self.train_val_split_seed = train_val_split_seed

        self.layout = layout
        self.static_layout = static_layout
        self.highresstatic_expand_t = highresstatic_expand_t
        self.mesostatic_expand_t = mesostatic_expand_t
        self.meso_crop = meso_crop
        self.fp16 = fp16
        # datamodule_only
        self.batch_size = batch_size
        self.num_workers = num_workers

    def prepare_data(self):
        assert os.path.exists(self.train_data_dir), "EarthNet2021 training set not found!"
        for test_data_dir in self.test_data_dir_list:
            assert os.path.exists(test_data_dir), f"EarthNet2021 test set at {test_data_dir} not found!"

    def setup(self, stage = None):
        if stage in (None, "fit"):
            train_val_data = EarthNet2021TrainDataset(
                return_mode=self.return_mode,
                data_aug_mode=self.data_aug_mode,
                data_aug_cfg=self.data_aug_cfg,
                data_dir=self.train_data_dir,
                layout=self.layout,
                static_layout=self.static_layout,
                highresstatic_expand_t=self.highresstatic_expand_t,
                mesostatic_expand_t=self.mesostatic_expand_t,
                meso_crop=self.meso_crop,
                fp16=self.fp16)
            val_size = int(self.val_ratio * len(train_val_data))
            train_size = len(train_val_data) - val_size

            if self.train_val_split_seed is not None:
                rnd_generator_dict = dict(generator=torch.Generator().manual_seed(self.train_val_split_seed))
            else:
                rnd_generator_dict = {}
            self.earthnet_train, self.earthnet_val = random_split(
                train_val_data, [train_size, val_size],
                **rnd_generator_dict)

        if stage in (None, "test"):
            self.earthnet_test_list = [
                EarthNet2021TestDataset(
                    return_mode=self.return_mode,
                    subset_name=test_subset_name,
                    data_dir=test_data_dir,
                    layout=self.layout,
                    static_layout=self.static_layout,
                    highresstatic_expand_t=self.highresstatic_expand_t,
                    mesostatic_expand_t=self.mesostatic_expand_t,
                    meso_crop=self.meso_crop,
                    fp16=self.fp16)
                for test_subset_name, test_data_dir in
                zip(self.test_subset_name_list, self.test_data_dir_list)]

        if stage in (None, "predict"):
            self.earthnet_predict_list = [
                EarthNet2021TestDataset(
                    return_mode=self.return_mode,
                    subset_name=test_subset_name,
                    data_dir=test_data_dir,
                    layout=self.layout,
                    static_layout=self.static_layout,
                    highresstatic_expand_t=self.highresstatic_expand_t,
                    mesostatic_expand_t=self.mesostatic_expand_t,
                    meso_crop=self.meso_crop,
                    fp16=self.fp16)
                for test_subset_name, test_data_dir in
                zip(self.test_subset_name_list, self.test_data_dir_list)]

    @property
    def num_train_samples(self):
        return len(self.earthnet_train)

    @property
    def num_val_samples(self):
        return len(self.earthnet_val)

    @property
    def num_test_samples(self):
        if len(self.earthnet_test_list) == 1:
            return len(self.earthnet_test_list[0])
        else:
            return [len(earthnet_test) for earthnet_test in self.earthnet_test_list]

    @property
    def num_predict_samples(self):
        if len(self.earthnet_predict_list) == 1:
            return len(self.earthnet_predict_list[0])
        else:
            return [len(earthnet_predict) for earthnet_predict in self.earthnet_predict_list]

    def train_dataloader(self):
        return DataLoader(self.earthnet_train, batch_size=self.batch_size, shuffle=True, num_workers=self.num_workers)

    def val_dataloader(self):
        return DataLoader(self.earthnet_val, batch_size=self.batch_size, shuffle=False, num_workers=self.num_workers)

    def test_dataloader(self):
        if len(self.earthnet_test_list) == 1:
            return DataLoader(self.earthnet_test_list[0], batch_size=self.batch_size, shuffle=False, num_workers=self.num_workers)
        else:
            return [DataLoader(earthnet_test, batch_size=self.batch_size, shuffle=False, num_workers=self.num_workers)
                    for earthnet_test in self.earthnet_test_list]

    def predict_dataloader(self):
        if len(self.earthnet_predict_list) == 1:
            return DataLoader(self.earthnet_predict_list[0], batch_size=self.batch_size, shuffle=False, num_workers=self.num_workers)
        else:
            return [
                DataLoader(earthnet_predict, batch_size=self.batch_size, shuffle=False, num_workers=self.num_workers)
                for earthnet_predict in self.earthnet_predict_list]

def get_EarthNet2021_dataloaders(
        dataloader_return_mode: str = "default",
        data_aug_mode: str = None,
        data_aug_cfg: Dict = None,
        train_data_dir: Union[Path, str] = None,
        test_subset_name: Union[str, Sequence[str]] = ("iid", "ood"),
        test_data_dir: Union[Union[Path, str], Sequence[Union[Path, str]]] = None,
        val_ratio: float = 0.1,
        train_val_split_seed: int = None,
        layout: str = None,
        static_layout: str = None,
        highresstatic_expand_t: bool = False,
        mesostatic_expand_t: bool = False,
        meso_crop: Union[str, Sequence[Sequence[int]]] = None,
        fp16: bool = False,
        batch_size=1,
        num_workers=8, ):

    if test_subset_name is None:
        if not isinstance(test_data_dir, Sequence):
            test_data_dir_list = [test_data_dir, ]
        else:
            test_data_dir_list = list(test_data_dir)
        test_subset_name_list = [None, ] * len(test_data_dir_list)
    else:
        if isinstance(test_subset_name, str):
            test_subset_name_list = [test_subset_name, ]
        elif isinstance(test_subset_name, Sequence):
            test_subset_name_list = list(test_subset_name)
        else:
            raise ValueError(f"Invalid type of test_subset_name {type(test_subset_name)}")
        test_data_dir_list = []
        for test_subset_name in test_subset_name_list:
            if test_subset_name == "iid":
                test_data_dir = EarthNet2021TestDataset.default_iid_test_data_dir
            elif test_subset_name == "ood":
                test_data_dir = EarthNet2021TestDataset.default_ood_test_data_dir
            elif test_subset_name == "extreme":
                test_data_dir = EarthNet2021TestDataset.default_extreme_test_data_dir
            elif test_subset_name == "seasonal":
                test_data_dir = EarthNet2021TestDataset.default_seasonal_test_data_dir
            else:
                raise ValueError(f"Invalid test_subset_name {test_subset_name}")
            test_data_dir_list.append(test_data_dir)

    train_val_data = EarthNet2021TrainDataset(
        return_mode=dataloader_return_mode,
        data_aug_mode=data_aug_mode,
        data_aug_cfg=data_aug_cfg,
        data_dir=train_data_dir,
        layout=layout,
        static_layout=static_layout,
        highresstatic_expand_t=highresstatic_expand_t,
        mesostatic_expand_t=mesostatic_expand_t,
        meso_crop=meso_crop,
        fp16=fp16)
    val_size = int(val_ratio * len(train_val_data))
    train_size = len(train_val_data) - val_size

    if train_val_split_seed is not None:
        rnd_generator_dict = dict(generator=torch.Generator().manual_seed(train_val_split_seed))
    else:
        rnd_generator_dict = {}
    earthnet_train, earthnet_val = random_split(
        train_val_data, [train_size, val_size],
        **rnd_generator_dict)

    earthnet_test_list = [
        EarthNet2021TestDataset(
            return_mode=dataloader_return_mode,
            subset_name=test_subset_name,
            data_dir=test_data_dir,
            layout=layout,
            static_layout=static_layout,
            highresstatic_expand_t=highresstatic_expand_t,
            mesostatic_expand_t=mesostatic_expand_t,
            meso_crop=meso_crop,
            fp16=fp16)
        for test_subset_name, test_data_dir in
        zip(test_subset_name_list, test_data_dir_list)]

    num_test_samples = [len(earthnet_test) for earthnet_test in earthnet_test_list]
    test_dataloader = [DataLoader(earthnet_test, batch_size=batch_size, shuffle=False, num_workers=num_workers)
                       for earthnet_test in earthnet_test_list]
    
    return {
        "train_dataloader": DataLoader(earthnet_train, batch_size=batch_size, shuffle=True, num_workers=num_workers),
        "val_dataloader": DataLoader(earthnet_val, batch_size=batch_size, shuffle=False, num_workers=num_workers),
        "test_dataloader": test_dataloader,
        "num_train_samples": len(earthnet_train),
        "num_val_samples": len(earthnet_val),
        "num_test_samples": num_test_samples,
    }
